import enum
from uuid import UUID

import pydantic

from neos_common import base

ResourcePattern = pydantic.constr(pattern=base.ResourceBase.RESOURCE_PATTERN)


class Statement(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(use_enum_values=True)

    sid: str
    principal: list[str] | UUID
    action: list[str]
    resource: list[ResourcePattern]  # type: ignore[reportGeneralTypeIssues]
    condition: list[str] = pydantic.Field(default_factory=list)
    effect: base.EffectEnum

    def is_allowed(self) -> bool:
        return self.effect == base.EffectEnum.allow.value


class PriorityStatement(Statement):
    priority: int


class Statements(pydantic.BaseModel):
    statements: list[Statement]


class PriorityStatements(pydantic.BaseModel):
    statements: list[PriorityStatement]


class PrincipalType(enum.Enum):
    model_config = pydantic.ConfigDict(use_enum_values=True)

    user = "user"
    group = "group"


class Principal(pydantic.BaseModel):
    principal_id: str
    principal_type: PrincipalType


class Principals(pydantic.BaseModel):
    principals: list[Principal]

    def get_principal_ids(self) -> list[str]:
        return [p.principal_id for p in self.principals]


class ErrorCode(pydantic.BaseModel):
    """Error code."""

    class_name: str
    type_: str = pydantic.Field(alias="type")
    title: str

    def model_dump(self, *args, **kwargs) -> dict:
        kwargs["by_alias"] = True
        return super().model_dump(*args, **kwargs)


class ErrorCodes(pydantic.BaseModel):
    """Error codes."""

    errors: list[ErrorCode]

    def model_dump(self, *args, **kwargs) -> dict:
        kwargs["by_alias"] = True
        return super().model_dump(*args, **kwargs)


class PermissionPair(pydantic.BaseModel):
    """Permission pair."""

    action: str
    resource: str


class FormattedRoute(pydantic.BaseModel):
    """Formatted route."""

    methods: str
    path: str
    permission_pairs: list[PermissionPair]
    summary: str | None = None
    logic_operator: str = "and"


class FormattedRoutes(pydantic.BaseModel):
    routes: list[FormattedRoute]



import abc
import datetime
import enum
import json
import logging
import random
import re
import string
import time
import typing
import uuid
from pathlib import Path

import pydantic

EMIT_LOG_FILENAME = "emit.log"
MAX_EMIT_LOG_LINE_LENGTH = 1000
ROOT_DIR = "/tmp/core-events"  # noqa: S108

logger = logging.getLogger(__name__)


class EventPacket(pydantic.BaseModel):
    source: str
    timestamp: int  # timestamp in ms
    span_id: pydantic.UUID4
    version: str
    message: str | dict[str, typing.Any]
    message_type: str

    @pydantic.field_validator("span_id")
    def span_id_to_str(cls, value: pydantic.UUID4) -> str:
        return str(value)


class EventPackets(pydantic.BaseModel):
    packets: list[EventPacket]


class EventVersion:
    VERSION_PATTERN = re.compile("^v(\\d+).(\\d+).(\\d+)$")

    def __init__(self, major: int, minor: int, patch: int) -> None:
        """Define version of the event.

        Args:
            major (int): version of the Gateway API
            minor (int): breaking changes
            patch (int): non-breaking changes
        """
        self.major = major
        self.minor = minor
        self.patch = patch

    def __eq__(self, other: object) -> bool:
        """Compare version for equality."""
        if not isinstance(other, EventVersion):
            return NotImplemented
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __le__(self, other: "EventVersion") -> bool:
        """Compare version."""
        return (self.major, self.minor, self.patch) <= (other.major, other.minor, other.patch)

    def __str__(self) -> str:
        """Convert to string."""
        return f"v{self.major}.{self.minor}.{self.patch}"

    @staticmethod
    def from_string(value: str) -> "EventVersion":
        m = EventVersion.VERSION_PATTERN.match(value)
        if not m:
            msg: str = f"Could not parse event version from {value}"
            raise ValueError(msg)
        return EventVersion(int(m.group(1)), int(m.group(2)), int(m.group(3)))


class EventType(enum.Enum):
    secret = "secret"  # noqa: S105
    tag = "tag"
    output = "output"
    data_product = "data_product"
    data_source = "data_source"
    data_system = "data_system"
    data_unit = "data_unit"
    spark = "spark"
    builder = "builder"
    journal_note = "journal_note"


class EventScope(enum.Enum):
    secret = EventType.secret.value
    secret_keys = f"{EventType.secret.value}.keys"

    tag = EventType.tag.value

    output = EventType.output.value
    output_info = f"{EventType.output.value}.info"
    output_journal = f"{EventType.output.value}.journal"

    data_product = EventType.data_product.value
    data_product_data = f"{EventType.data_product.value}.data"
    data_product_metadata = f"{EventType.data_product.value}.metadata"
    data_product_output = f"{EventType.data_product.value}.output"
    data_product_data_product = f"{EventType.data_product.value}.data_product"
    data_product_info = f"{EventType.data_product.value}.info"
    data_product_journal = f"{EventType.data_product.value}.journal"
    data_product_schema = f"{EventType.data_product.value}.schema"
    data_product_spark = f"{EventType.data_product.value}.spark"
    data_product_spark_state = f"{EventType.data_product.value}.spark_state"
    data_product_spark_builder = f"{EventType.data_product.value}.spark_builder"
    data_product_expectations = f"{EventType.data_product.value}.expectations"
    data_product_data_quality = f"{EventType.data_product.value}.data_quality"
    data_product_profiling = f"{EventType.data_product.value}.profiling"
    data_product_validation = f"{EventType.data_product.value}.validation"
    data_product_classification = f"{EventType.data_product.value}.classification"
    data_product_classification_configuration = f"{EventType.data_product.value}.classification.configuration"
    data_product_classification_result = f"{EventType.data_product.value}.classification.result"
    data_product_health = f"{EventType.data_product.value}.health"

    data_source = EventType.data_source.value
    data_source_info = f"{EventType.data_source.value}.info"
    data_source_journal = f"{EventType.data_source.value}.journal"
    data_source_connection = f"{EventType.data_source.value}.connection"
    data_source_health = f"{EventType.data_source.value}.health"
    data_source_secret = f"{EventType.data_source.value}.secret"
    data_source_state = f"{EventType.data_source.value}.state"
    data_source_data_unit = f"{EventType.data_source.value}.data_unit"

    data_system = EventType.data_system.value
    data_system_info = f"{EventType.data_system.value}.info"
    data_system_journal = f"{EventType.data_system.value}.journal"
    data_system_data_source = f"{EventType.data_system.value}.data_source"

    data_unit = EventType.data_unit.value
    data_unit_metadata = f"{EventType.data_unit.value}.metadata"
    data_unit_schema = f"{EventType.data_unit.value}.schema"
    data_unit_state = f"{EventType.data_unit.value}.state"
    data_unit_info = f"{EventType.data_unit.value}.info"
    data_unit_journal = f"{EventType.data_unit.value}.journal"
    data_unit_health = f"{EventType.data_unit.value}.health"
    data_unit_profiling = f"{EventType.data_unit.value}.profiling"
    data_unit_config = f"{EventType.data_unit.value}.config"
    data_unit_data_product = f"{EventType.data_unit.value}.data_product"

    spark = EventType.spark.value
    spark_status = f"{EventType.spark.value}.status"

    builder = EventType.builder.value

    journal_note = EventType.journal_note.value


class EventAction(enum.Enum):
    secret_add = f"{EventScope.secret.value}:add"
    secret_update = f"{EventScope.secret.value}:update"
    secret_delete = f"{EventScope.secret.value}:delete"
    secret_keys_delete = f"{EventScope.secret_keys.value}:delete"

    tag_add = f"{EventScope.tag.value}:add"
    tag_delete = f"{EventScope.tag.value}:delete"

    output_create = f"{EventScope.output.value}:create"
    output_update = f"{EventScope.output.value}:update"
    output_delete = f"{EventScope.output.value}:delete"
    output_info_update = f"{EventScope.output_info.value}:update"
    output_journal_add = f"{EventScope.output_journal.value}:add"

    data_product_create = f"{EventScope.data_product.value}:create"
    data_product_update = f"{EventScope.data_product.value}:update"
    data_product_delete = f"{EventScope.data_product.value}:delete"
    data_product_data_delete = f"{EventScope.data_product_data.value}:delete"
    data_product_metadata_update = f"{EventScope.data_product_metadata.value}:update"
    data_product_metadata_remove = f"{EventScope.data_product_metadata.value}:remove"
    data_product_schema_update = f"{EventScope.data_product_schema.value}:update"
    data_product_output_link = f"{EventScope.data_product_output.value}:link"
    data_product_output_unlink = f"{EventScope.data_product_output.value}:unlink"
    data_product_data_product_link = f"{EventScope.data_product_data_product.value}:link"
    data_product_data_product_unlink = f"{EventScope.data_product_data_product.value}:unlink"
    data_product_info_update = f"{EventScope.data_product_info.value}:update"
    data_product_journal_add = f"{EventScope.data_product_journal.value}:add"
    data_product_publish = f"{EventScope.data_product.value}:publish"
    data_product_unpublish = f"{EventScope.data_product.value}:unpublish"
    data_product_spark_update = f"{EventScope.data_product_spark.value}:update"
    data_product_spark_builder_update = f"{EventScope.data_product_spark_builder.value}:update"
    data_product_spark_state_update = f"{EventScope.data_product_spark_state.value}:update"
    data_product_expectations_update = f"{EventScope.data_product_expectations.value}:update"
    data_product_profiling_update = f"{EventScope.data_product_profiling.value}:update"
    data_product_validation_update = f"{EventScope.data_product_validation.value}:update"
    data_product_classification_update = f"{EventScope.data_product_classification.value}:update"
    data_product_classification_configuration_update = (
        f"{EventScope.data_product_classification_configuration.value}:update"
    )
    data_product_classification_result_update = f"{EventScope.data_product_classification_result.value}:update"
    data_product_health_check = f"{EventScope.data_product_health.value}:check"

    data_source_create = f"{EventScope.data_source.value}:create"
    data_source_update = f"{EventScope.data_source.value}:update"
    data_source_delete = f"{EventScope.data_source.value}:delete"
    data_source_info_update = f"{EventScope.data_source_info.value}:update"
    data_source_journal_add = f"{EventScope.data_source_journal.value}:add"
    data_source_connection_update = f"{EventScope.data_source_connection.value}:update"
    data_source_secret_add = f"{EventScope.data_source_secret.value}:add"
    data_source_data_unit_link = f"{EventScope.data_source_data_unit.value}:link"
    data_source_data_unit_unlink = f"{EventScope.data_source_data_unit.value}:unlink"
    data_source_health_check = f"{EventScope.data_source_health.value}:check"
    data_source_state_update = f"{EventScope.data_source_state.value}:update"

    data_system_create = f"{EventScope.data_system.value}:create"
    data_system_update = f"{EventScope.data_system.value}:update"
    data_system_delete = f"{EventScope.data_system.value}:delete"
    data_system_info_update = f"{EventScope.data_system_info.value}:update"
    data_system_info_add = f"{EventScope.data_system_info.value}:add"
    data_system_journal_add = f"{EventScope.data_system_journal.value}:add"
    data_system_data_source_link = f"{EventScope.data_system_data_source.value}:link"
    data_system_data_source_unlink = f"{EventScope.data_system_data_source.value}:unlink"

    data_unit_create = f"{EventScope.data_unit.value}:create"
    data_unit_update = f"{EventScope.data_unit.value}:update"
    data_unit_metadata_update = f"{EventScope.data_unit_metadata.value}:update"
    data_unit_metadata_delete = f"{EventScope.data_unit_metadata.value}:delete"
    data_unit_schema_update = f"{EventScope.data_unit_schema.value}:update"
    data_unit_state_update = f"{EventScope.data_unit_state.value}:update"
    data_unit_delete = f"{EventScope.data_unit.value}:delete"
    data_unit_info_update = f"{EventScope.data_unit_info.value}:update"
    data_unit_journal_add = f"{EventScope.data_unit_journal.value}:add"
    data_unit_config_update = f"{EventScope.data_unit_config.value}:update"
    data_unit_data_product_link = f"{EventScope.data_unit_data_product.value}:link"
    data_unit_data_product_unlink = f"{EventScope.data_unit_data_product.value}:unlink"
    data_unit_health_check = f"{EventScope.data_unit_health.value}:check"
    data_unit_profiling_update = f"{EventScope.data_unit_profiling.value}:update"

    spark_finish = f"{EventScope.spark.value}:finish"
    spark_status_update = f"{EventScope.spark_status.value}:update"

    builder_preview = f"{EventScope.builder.value}:preview"

    journal_note_update = f"{EventScope.journal_note.value}:update"
    journal_note_delete = f"{EventScope.journal_note.value}:delete"


class Emitter(abc.ABC):
    @abc.abstractmethod
    async def emit(self, event_packet: EventPacket) -> None:
        pass


class SyncEmitter(abc.ABC):
    @abc.abstractmethod
    def emit(self, event_packet: EventPacket) -> None:
        pass


class TmpEmitter(Emitter):
    def __init__(self, max_emit_log_line_length: int = MAX_EMIT_LOG_LINE_LENGTH) -> None:
        self.max_emit_log_line_length = max_emit_log_line_length
        self.root_dir = self.ensure_root_folder_exists()

    @staticmethod
    def split_uuid_to_dirs_and_filename(span_id: uuid.UUID) -> tuple[str, str, str]:
        """Split span_id (uuid) into nested directories.

        Helper function that splits span_id (uuid) into nested directories.

        span_dir  span_subdir  span_filename
           87   /      39    / 0265-e23d-4574-b71e-dd1d61a74496

        for the purpose of storing large amount of files and maintaining performance
        with filesystems and other os tools

        Args:
        ----
        span_id: uuid.uuid4

        Returns: tuple (span_dir, span_subdir, span_filename)
        """
        span_id_str = str(span_id)

        span_dir = span_id_str[0:2]
        span_subdir = span_id_str[2:4]
        span_filename = span_id_str[4:]

        return span_dir, span_subdir, span_filename

    @staticmethod
    def ensure_root_folder_exists() -> Path:
        root_dir = Path(ROOT_DIR)
        root_dir.mkdir(exist_ok=True)

        return root_dir

    def get_event_filepath(self, span_id: uuid.UUID) -> Path:
        span_dir, span_subdir, span_filename = self.split_uuid_to_dirs_and_filename(span_id)

        return self.root_dir / span_dir / span_subdir / span_filename

    def read_event_packets(self, span_id: uuid.UUID) -> EventPackets:
        event_filepath = self.get_event_filepath(span_id)

        if event_filepath.is_file():
            with event_filepath.open("r") as f:
                packets = json.load(f)
                event_packets = EventPackets(**packets)
        else:
            event_packets = EventPackets(packets=[])

        return event_packets

    def save_event_packets(self, span_id: uuid.UUID, event_packets: EventPackets) -> None:
        event_filepath = self.get_event_filepath(span_id)

        with event_filepath.open("w") as f:
            json.dump(event_packets.model_dump(), f)

    def clear_oldest_span_id_event(self, span_id: uuid.UUID) -> None:
        # Remove span_id file
        dir_0, dir_1, filename = self.split_uuid_to_dirs_and_filename(span_id)

        # empty oldest info in span file
        event_packets = self.read_event_packets(span_id)
        if event_packets.packets:
            event_packets.packets.pop(0)

        if not event_packets.packets:
            # remove file
            self.get_event_filepath(span_id).unlink(missing_ok=True)

            # Remove span subdir if empty
            span_sub_dir = self.root_dir / dir_0 / dir_1
            file_count = len([child for child in span_sub_dir.iterdir() if child.is_file()])
            if file_count == 0:
                span_sub_dir.rmdir()

            # Remove span dir if empty
            span_dir = self.root_dir / dir_0
            dir_count = len([child for child in span_dir.iterdir() if child.is_dir()])
            if dir_count == 0:
                span_dir.rmdir()
        else:
            self.save_event_packets(span_id, event_packets)

    def ensure_emit_log_keep_size(self) -> None:
        history_log = self.root_dir / EMIT_LOG_FILENAME

        random_str = "".join(random.choice(string.ascii_lowercase) for i in range(8))  # noqa: S311
        history_log_tmp = self.root_dir / f"{random_str}_{EMIT_LOG_FILENAME}.tmp"

        if history_log.is_file():
            with history_log.open("r") as f:
                number_of_lines = len(f.readlines())

            diff = number_of_lines - (self.max_emit_log_line_length - 1)

            if diff > 0:
                with history_log.open("r") as original_f:
                    # skip the lines
                    for _i in range(diff):
                        # Remove line from event_log by moving the file pointer
                        line = original_f.readline().rstrip("\n")

                        span_id = uuid.UUID(line.split(" ")[-1])
                        self.clear_oldest_span_id_event(span_id)

                    # copy the rest
                    with history_log_tmp.open("w") as tmp_f:
                        for line in original_f:
                            tmp_f.write(line)

                # remove original and rename tmp file
                history_log.unlink()
                history_log_tmp.rename(history_log)

    def update_emit_log(self, event_packet: EventPacket) -> None:
        history_log = self.root_dir / EMIT_LOG_FILENAME

        self.ensure_emit_log_keep_size()

        with history_log.open("a") as f:
            f.write(
                f"{datetime.datetime.fromtimestamp(event_packet.timestamp / 1000.0, tz=datetime.timezone.utc)} {event_packet.message_type} {event_packet.span_id!s}\n",  # noqa: E501
            )

    async def emit(self, event_packet: EventPacket) -> None:
        try:
            dir_0, dir_1, filename = self.split_uuid_to_dirs_and_filename(event_packet.span_id)
            directory = self.root_dir / dir_0 / dir_1
            directory.mkdir(parents=True, exist_ok=True)

            event_packets = self.read_event_packets(event_packet.span_id)
            event_packets.packets.append(event_packet)

            self.save_event_packets(event_packet.span_id, event_packets)

            self.update_emit_log(event_packet)
        except Exception:
            logger.exception("Failed to emit event.")


def _prepare_event_packet(
    message: str | typing.Any,  # noqa: ANN401
    source: str,
    version: EventVersion,
    span_id: uuid.UUID | None = None,
    message_type: str = "unknown",
) -> EventPacket:
    span_id = span_id or uuid.uuid4()

    try:
        deserialized_message = json.loads(message)
    except (json.JSONDecodeError, TypeError):
        deserialized_message = message

    return EventPacket(
        source=source,
        timestamp=int(time.time() * 1000),  # timestamp in ms
        span_id=span_id,
        version=str(version),
        message=deserialized_message,
        message_type=message_type,
    )


class Event(abc.ABC):
    def __init__(self, emitter: Emitter, source: str = "source") -> None:
        self.emitter = emitter
        self.source = source

    @abc.abstractmethod
    def version(self) -> EventVersion: ...

    async def emit(
        self,
        message: str,
        span_id: uuid.UUID | None = None,
        message_type: str = "unknown",
    ) -> uuid.UUID:
        span_id = span_id or uuid.uuid4()

        event_packet = _prepare_event_packet(
            message=message,
            source=self.source,
            span_id=span_id,
            version=self.version(),
            message_type=message_type,
        )

        await self.emitter.emit(event_packet)

        return span_id


class SyncEvent(abc.ABC):
    @abc.abstractmethod
    def version(self) -> EventVersion: ...

    def __init__(self, emitter: SyncEmitter, source: str = "source") -> None:
        self.emitter = emitter
        self.source = source

    def emit(
        self,
        message: str,
        span_id: uuid.UUID | None = None,
        message_type: str = "unknown",
    ) -> uuid.UUID:
        span_id = span_id or uuid.uuid4()

        event_packet = _prepare_event_packet(
            message=message,
            source=self.source,
            span_id=span_id,
            version=self.version(),
            message_type=message_type,
        )

        self.emitter.emit(event_packet)

        return span_id


class ValidationResult(pydantic.BaseModel):
    """Data Quality validation results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    entity_identifier: uuid.UUID
    expectations_id: uuid.UUID
    success: bool
    success_percentage: float
    score: int
    column_score: dict[str, int]
    category_weights: dict
    threshold: float
    field_thresholds: dict
    raw_result: dict
    global_stats: dict
    table_stats: dict
    column_stats: dict


class ProfilingResult(pydantic.BaseModel):
    """Profiling results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    created_at: datetime.datetime
    entity_identifier: uuid.UUID
    dataset_type: str
    row_count: int
    column_count: int
    profile_sample: float
    column_stats: list[dict] | dict[str, typing.Any]


class ClassificationResult(pydantic.BaseModel):
    """Data classification results model."""

    id_: uuid.UUID = pydantic.Field(alias="id")
    entity_identifier: uuid.UUID
    dataset_type: str
    config: dict
    column_results: dict



"""Common utils to support neos cli applications."""

import typing
from csv import DictWriter
from io import StringIO

from fastapi import FastAPI

from neos_common.util import get_routes


class RoutesPrinter:
    """Routes printer."""

    def __init__(self, app: FastAPI, ignore_routes: list[str]) -> None:
        self.routes = get_routes(app, ignore_routes)

    def echo(self, echo_fn: typing.Callable) -> None:
        """Print routes to console."""
        for route in self.routes:
            echo_fn(f"{route.summary}")
            echo_fn(f"{route.methods:<10}{route.path}")
            if route.permission_pairs:
                echo_fn("{:<12}{}".format("", route.logic_operator.upper()))
            for permission_pair in route.permission_pairs:
                echo_fn("{:<14}action:   {}".format("", permission_pair.action))
                echo_fn("{:<14}resource: {}".format("", permission_pair.resource))
                echo_fn()
            echo_fn()


class RoutesCSVPrinter(RoutesPrinter):
    """Routes CSV printer."""

    def echo(self, echo_fn: typing.Callable) -> None:
        """Print routes in csv format."""
        csv_output = StringIO()
        writer = DictWriter(csv_output, fieldnames=["methods", "path", "summary", "operator", "action", "resource"])
        writer.writeheader()

        for route in self.routes:
            header = True
            if len(route.permission_pairs) == 0:
                writer.writerow(
                    {
                        "methods": route.methods,
                        "path": route.path,
                        "summary": route.summary,
                        "operator": "",
                        "action": "",
                        "resource": "",
                    },
                )
            else:
                for permission_pair in route.permission_pairs:
                    writer.writerow(
                        {
                            "methods": route.methods if header else "",
                            "path": route.path if header else "",
                            "summary": route.summary,
                            "operator": route.logic_operator.upper(),
                            "action": permission_pair.action,
                            "resource": permission_pair.resource,
                        },
                    )
                    header = False

        csv_output.seek(0)
        echo_fn(csv_output.read())



import re
import typing as t
from http import HTTPStatus

CONVERT_RE = re.compile(r"(?<!^)(?=[A-Z])")

if t.TYPE_CHECKING:
    Loc = tuple[int | str, ...]

    class _ErrorDictRequired(t.TypedDict):
        loc: Loc
        msg: str
        type: str

    class ErrorDict(_ErrorDictRequired, total=False):
        """Type hint for pydantic.error_wrappers.ErrorDict.

        Stolen from type hints in pydantic.error_wrappers
        """

        ctx: dict[str, t.Any]


class NeosException(Exception):  # noqa: N818
    """Base class for all neos exceptions."""

    status_ = "unhandled"
    reason_ = None
    message_ = ""

    def __init__(
        self,
        debug: str | None = None,
        reason: str | None = None,
        message: str | None = None,
        status: int | str | None = None,
        details: list["ErrorDict"] | None = None,
    ) -> None:
        """Socket exception initiator.

        Args:
        ----
        debug: Additional context for specific error instance
        reason: error reason i.e. "neos-exception"
        status: error status i.e. "not-found"
        message: error message
        details: optional list of pydantic validation errors
        """
        reason_ = self.__class__.__name__.replace("Error", "")
        reason_ = CONVERT_RE.sub("-", reason_).lower()

        if isinstance(status, int):
            status = HTTPStatus(status).phrase.replace(" ", "-").lower()

        self.reason = reason or self.reason_ or reason_
        self.message = message or self.message_
        self.status = status or self.status_
        self.debug = debug
        self.details = details

    def __iter__(self) -> t.Iterator[tuple[str, bool | str | list["ErrorDict"] | None]]:
        """Return an iterator to support `dict(self)`."""
        return iter(
            [
                ("ok", False),
                ("status", self.status),
                ("reason", self.reason),
                ("message", self.message),
                ("debug", self.debug),
                ("details", self.details),
            ],
        )


class UnauthorisedError(NeosException):
    status_ = "unauthorised"


class ForbiddenError(NeosException):
    status_ = "forbidden"


class NotFoundError(NeosException):
    status_ = "not-found"


class ConflictError(NeosException):
    status_ = "conflict"


class BadRequestError(NeosException):
    status_ = "bad-request"


class ServerError(NeosException):
    status_ = "server-error"


class UnhandledServiceApiError(NeosException):
    pass


class ServiceApiError(NeosException):
    pass


class AuthorizationRequiredError(UnauthorisedError):
    message_ = "Authorization token required."


class InvalidAuthorizationError(UnauthorisedError):
    message_ = "Authorization token invalid."


class InsufficientPermissionsError(ForbiddenError):
    message_ = "Insufficient permissions."


class InvalidResourceFormatError(BadRequestError):
    message_ = "Resource has invalid format."


class IdentityAccessManagerError(NeosException):
    reason_ = "identity-access-manager-error"
    message_ = "Problem with Identity Access Manager."


class ServiceConnectionError(NeosException):
    reason_ = "service-connection-error"

    def __init__(self, message: str, debug: str | None = None) -> None:
        super().__init__(message=message, debug=debug)


class ServiceTimeoutError(NeosException):
    reason_ = "service-timeout-error"

    def __init__(self, message: str, debug: str | None = None) -> None:
        super().__init__(message=message, debug=debug)


class SignatureError(UnauthorisedError):
    reason_ = "signature-error"
    message_ = "Signature invalid."



import contextlib

from fastapi import FastAPI

from neos_common.schema import FormattedRoute, PermissionPair


def get_routes(app: FastAPI, ignore_routes: list[str]) -> list[FormattedRoute]:
    routes = []
    for route in app.routes:
        if route.path in ignore_routes:
            continue

        methods = ""
        if hasattr(route, "methods"):
            methods = ", ".join(sorted(route.methods))

        path = route.path
        summary = None
        with contextlib.suppress(AttributeError):
            summary = route.summary

        permission_pairs = []
        logic_operator = "and"
        if hasattr(route, "openapi_extra") and route.openapi_extra is not None:
            actions = []
            resources = []
            for key in sorted(route.openapi_extra):
                if key.startswith("x-iam-action"):
                    actions.append(route.openapi_extra[key])
                if key.startswith("x-iam-resource"):
                    resources.append(route.openapi_extra[key])

            permission_pairs = [PermissionPair(action=a, resource=r) for a, r in zip(actions, resources, strict=False)]
            logic_operator = route.openapi_extra.get("logic-operator", "and")

        routes.append(
            FormattedRoute(
                methods=methods,
                path=path,
                permission_pairs=permission_pairs,
                summary=summary,
                logic_operator=logic_operator,
            ),
        )

    return routes



import typing as t

import dotenv
import pydantic
import pydantic_settings

T = t.TypeVar("T", bound=pydantic.BaseModel)


class Config(pydantic_settings.BaseSettings):
    """Service configuration base class."""

    @classmethod
    def load(cls: type[T]) -> T:
        env = dotenv.find_dotenv(usecwd=True)

        return cls(_env_file=env)

    model_config = pydantic_settings.SettingsConfigDict(
        env_prefix="nc_",
        extra="ignore",
    )

    name: str = "common"

    logging_level: str = pydantic.Field(default="INFO")

    # -- DATABASE --

    postgres_host: str
    postgres_user: str
    postgres_database: str
    postgres_password: str
    postgres_port: int = pydantic.Field(default=5432)

    postgres_pool_min_size: int = pydantic.Field(default=1)
    postgres_pool_max_size: int = pydantic.Field(default=1)

    @property
    def postgres_dsn(self) -> str:
        """Generate postgres dsn from provided configuration."""
        auth = f"{self.postgres_user}:{self.postgres_password}"
        host = f"{self.postgres_host}:{self.postgres_port}"
        return f"postgresql://{auth}@{host}/{self.postgres_database}"

    @property
    def logging_config(self) -> dict:
        """Generate default logging config."""
        return {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "level": self.logging_level.upper(),
                    "formatter": "default",
                },
            },
            "loggers": {
                self.name: {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
                "pogo_migrate": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
            },
        }


class ApiConfig(Config):
    """API Service configuration base class."""

    name: str = "common-api"

    raw_api_prefix: str = pydantic.Field(default="", alias="NC_API_PREFIX")

    @property
    def api_prefix(self) -> str:
        """Ensure api prefix starts with /."""
        api_prefix = self.raw_api_prefix
        if api_prefix and not api_prefix.startswith("/"):
            api_prefix = f"/{api_prefix}"
        return api_prefix

    # -- KEYCLOAK --

    keycloak_host: str
    keycloak_realm: str
    keycloak_client_id: str
    keycloak_client_secret: str

    # -- ACCESS SECRET --
    access_key_id: str
    secret_access_key: str
    partition: str = pydantic.Field(default="ksa")

    @property
    def logging_config(self) -> dict:
        """Generate default logging config including api loggers."""
        return {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "format": "[%(asctime)s][%(name)s][%(levelname)s]: %(message)s",
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                },
            },
            "handlers": {
                "default": {
                    "class": "logging.StreamHandler",
                    "level": self.logging_level.upper(),
                    "formatter": "default",
                },
            },
            "loggers": {
                self.name: {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
                "uvicorn.access": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "pogo_migrate": {
                    "handlers": ["default"],
                    "level": "INFO",
                    "propagate": False,
                },
                "neos_common": {
                    "handlers": ["default"],
                    "level": self.logging_level.upper(),
                    "propagate": False,
                },
            },
        }



import asyncio
import json
import logging
import socketserver
import typing

import pydantic
import pydantic.error_wrappers

from neos_common.error import NeosException
from neos_common.socket import util

if typing.TYPE_CHECKING:
    Loc = tuple[int | str, ...]

    class _ErrorDictRequired(typing.TypedDict):
        loc: Loc
        msg: str
        type: str

    class ErrorDict(_ErrorDictRequired, total=False):
        """Type hint for pydantic.error_wrappers.ErrorDict.

        Stolen from type hints in pydantic.error_wrappers
        """

        ctx: dict[str, typing.Any]


logger = logging.getLogger(__name__)


class ValidationError(NeosException):
    status_ = "unprocessible-entity"


class AsyncTCPHandler:
    Request = util.SocketRequest

    async def setup(self) -> None: ...

    async def teardown(self) -> None: ...

    async def __call__(self, reader: asyncio.streams.StreamReader, writer: asyncio.streams.StreamWriter) -> None:
        async_socket = util.AsyncSocket(reader, writer)
        await self.setup()
        try:
            data = await async_socket.read()
        except json.decoder.JSONDecodeError as e:
            logger.exception("Decode error processing request.")
            exc = ValidationError(
                reason="DecodeError",
                message=str(e),
                details=None,
            )
            async_socket.write(util.encode(dict(exc)))
        except Exception as e:
            logger.exception("Exception processing request.")
            exc = NeosException(
                reason="Unhandled",
                message=str(e),
            )
            async_socket.write(util.encode(dict(exc)))
        else:
            if data is None:
                await self.teardown()
                async_socket.close()
                return

            try:
                request = self.Request(**data)
                request_type = request.request_type

                RequestDataCls = self.request_mapping.get(request_type, dict)  # noqa: N806
                request_data = RequestDataCls(**request.data)
            except pydantic.ValidationError as e:
                logger.exception("Validation error processing request data.")
                exc = ValidationError(
                    reason="ValidationError",
                    message="Unable to parse message.",
                    details=e.errors(),
                )
                async_socket.write(util.encode(dict(exc)))
            except Exception as e:
                logger.exception("Exception processing request data.")
                exc = NeosException(
                    reason="Unhandled",
                    message=str(e),
                )
                async_socket.write(util.encode(dict(exc)))
            else:
                await self._handle(async_socket, request_type, request_data)

        await self.teardown()
        async_socket.close()

    async def _handle(
        self,
        sock: util.AsyncSocket,
        request_type: str,
        request_data: pydantic.BaseModel | dict,
    ) -> None: ...


class TCPHandler(socketserver.BaseRequestHandler):
    """Base handler for TCP socket server.

    When a message is received:
        * decode it
        * format it into an instance of `cls.Request`
        * pass it to the handler defined for `request.request_type` in `cls.request_mapping`

    Define the mapping between Request.request_typpe and handlers in `cls.request_mapping`.
    Override the default SocketRequest schema via `cls.Request`
    """

    request_mapping: typing.ClassVar[dict] = {}
    Request = util.SocketRequest

    def handle(self) -> None:
        """Handle an incoming request.

        Handle issues decoding, processing and validating request messages, on
        error response to the client, with error details.

        On success pass the validated request message to the appropriate
        handler.
        """
        try:
            data = util.decode(util.recv_msg(self.request))
        except json.decoder.JSONDecodeError as e:
            logger.exception("Decode error processing request.")
            exc = ValidationError(
                reason="DecodeError",
                message=str(e),
                details=None,
            )
            util.send_msg(self.request, util.encode(dict(exc)))
        except Exception as e:
            logger.exception("Exception processing request.")
            exc = NeosException(
                reason="Unhandled",
                message=str(e),
            )
            util.send_msg(self.request, util.encode(dict(exc)))
        else:
            if data is None:
                return

            try:
                request = self.Request(**data)
                request_type = request.request_type

                RequestDataCls = self.request_mapping.get(request_type, dict)  # noqa: N806
                request_data = RequestDataCls(**request.data)
            except pydantic.ValidationError as e:
                logger.exception("Validation error processing request data.")
                exc = ValidationError(
                    reason="ValidationError",
                    message="Unable to parse message.",
                    details=e.errors(),
                )
                util.send_msg(self.request, util.encode(dict(exc)))
            except Exception as e:
                logger.exception("Exception processing request data.")
                exc = NeosException(
                    reason="Unhandled",
                    message=str(e),
                )
                util.send_msg(self.request, util.encode(dict(exc)))
            else:
                self._handle(self.request, request_type, request_data)

    def _handle(
        self,
        sock: util.Socket,
        request_type: str,
        request_data: pydantic.BaseModel | dict,
    ) -> None: ...



import logging
import uuid

import fastapi

from neos_common.authorization import base, signer
from neos_common.base import Action, ResourceLike
from neos_common.client.hub_client import HubClient

logger = logging.getLogger(__name__)


class AccessValidator(base.AccessValidator):
    def __init__(self, hub_client: HubClient) -> None:
        self._hub_client = hub_client

    async def validate(
        self,
        user_id: uuid.UUID,
        actions: list[Action],
        resources: list[ResourceLike],
        logic_operator: str,
        *,
        return_allowed_resources: bool = False,
    ) -> tuple[uuid.UUID, list[str]]:
        return await self._hub_client.validate_token(
            principal=user_id,
            actions=actions,
            resources=[resource.urn for resource in resources],
            logic_operator=logic_operator,
            return_allowed_resources=return_allowed_resources,
        )


class SignatureValidator(base.SignatureValidator):
    def __init__(self, hub_client: HubClient) -> None:
        self._hub_client = hub_client

    async def validate(
        self,
        request: fastapi.Request,
        actions: list[Action],
        resources: list[ResourceLike],
        logic_operator: str,
        *,
        return_allowed_resources: bool = False,
    ) -> tuple[uuid.UUID, list[str]]:
        validator = signer.Validator()

        payload = await request.body()

        auth_type, access_key_id, scope, challenge, signature = validator.challenge_v4(
            request.method,
            request.url,
            request.headers,
            payload,
        )

        return await self._hub_client.validate_signature(
            access_key_id,
            auth_type.split("-")[0],
            scope,
            challenge,
            signature,
            actions,
            [resource.urn for resource in resources],
            logic_operator=logic_operator,
            return_allowed_resources=return_allowed_resources,
        )



import asyncio
import json
import socket
import typing

import pydantic

Socket = socket.socket


class SocketRequest(pydantic.BaseModel):
    request_type: str
    data: dict[str, typing.Any]


def encode(obj: dict, encoder: type[json.JSONEncoder] | None = None) -> bytes:
    """JSON encode message object for sending via socket."""
    new = json.dumps(obj, cls=encoder)
    return new.encode()


def decode(obj: bytes | None) -> dict[str, typing.Any] | None:
    """Decode message received via socket."""
    if obj is None:
        return obj

    data: dict[str, typing.Any] = json.loads(obj.decode())
    return data


class AsyncSocket:
    def __init__(self, reader: asyncio.streams.StreamReader, writer: asyncio.streams.StreamWriter) -> None:
        self.reader = reader
        self.writer = writer

    def write(self, msg: bytes) -> None:
        """Send a message via the socket."""
        msg = add_msg_header(msg)
        self.writer.write(msg)

    async def read(self) -> dict | None:
        """Receive a message via the socket."""
        # start by getting the header
        # (which is an int of length `BYTE_COUNT`).
        # The header tells the message size in bytes.
        raw_msglen = await self.reader.read(BYTE_COUNT)
        # Then retrieve a message of length `raw_msglen`
        # this will be the actual message
        msglen = len_frombytes(raw_msglen)

        data = bytearray()
        while len(data) < msglen:
            packet = await self.reader.read(msglen - len(data))
            if not packet:
                return None
            data.extend(packet)

        return decode(data)

    def close(self) -> None:
        self.writer.close()


BYTE_COUNT = 4


def send_msg(sock: Socket, msg: bytes) -> None:
    """Send a message via the socket."""
    msg = add_msg_header(msg)
    sock.sendall(msg)


def recv_msg(sock: Socket) -> bytearray | None:
    """Receive a message via the socket."""
    # start by getting the header
    # (which is an int of length `BYTE_COUNT`).
    # The header tells the message size in bytes.
    raw_msglen = recvall(sock, BYTE_COUNT)
    if not raw_msglen:
        return None
    # Then retrieve a message of length `raw_msglen`
    # this will be the actual message
    msglen = len_frombytes(raw_msglen)
    return recvall(sock, msglen)


def recvall(sock: Socket, length: int) -> bytearray | None:
    """Get a message of a certain length from the socket stream."""
    data = bytearray()
    while len(data) < length:
        packet = sock.recv(length - len(data))
        if not packet:
            return None
        data.extend(packet)
    return data


def add_msg_header(msg: bytes) -> bytes:
    """Prepend message length header onto message.

    Args:
    ----
    msg: message bytestring

    Returns:
    -------
    new bytestring with original message length prepended.
    """
    return len_inbytes(msg) + msg


def len_inbytes(msg: bytes) -> bytes:
    """Retrieve length of message as a bytestring."""
    return len(msg).to_bytes(BYTE_COUNT, byteorder="big")


def len_frombytes(bmsg: bytes) -> int:
    """Extract length of message from a bytestring header."""
    return int.from_bytes(bmsg, byteorder="big")



import logging
import time
from functools import partial

from confluent_kafka import KafkaException, Producer, cimpl


class KafkaClient:
    def __init__(
        self,
        host: str,
        logger_name: str,
        timing_logger_name: str,
        poll_timeout: float,
    ) -> None:
        self.host = host
        self.logger = logging.getLogger(logger_name)
        self.timing_logger = logging.getLogger(timing_logger_name)
        self.poll_timeout = poll_timeout
        self._producer = Producer(self._render_config(self.host))

    def _render_config(self, host: str) -> dict:
        return {
            "bootstrap.servers": host,
        }

    def _ack(
        self,
        start: float,
        topic: str,
        err: KafkaException,
        _: cimpl.Message,
    ) -> None:
        duration = (time.time() - start) * 1000

        if err:
            self.logger.exception(str(err))
            self.timing_logger.info(
                f"TIMING: Wall: {duration:.2f}ms | {err!s}",
            )
        else:
            self.logger.info(f"Kafka produce to topic: {topic}")
            self.timing_logger.info(
                f"TIMING: Wall: {duration:.2f}ms | Kafka produce to topic: {topic}",
            )

    def produce(
        self,
        topic: str,
        value: str | bytes,
        key: str | bytes | None = None,
        partition: int | None = None,
        timestamp: int | None = None,
        headers: dict | list | None = None,
    ) -> None:
        """Produce message to the Kafka topic."""
        start = time.time()

        kwargs = {
            k: v
            for k, v in {
                "key": key,
                "partition": partition,
                "on_delivery": partial(self._ack, start, topic),
                "timestamp": timestamp,
                "headers": headers,
            }.items()
            if v is not None
        }

        self._producer.produce(topic, value, **kwargs)
        self._producer.poll(self.poll_timeout)



import abc
import re
import typing as t
from dataclasses import dataclass
from enum import Enum

import pydantic

from neos_common import error

T = t.TypeVar("T", bound=pydantic.BaseModel)


STAR = "*"
ROOT = "root"


class Partition(Enum):
    ksa = "ksa"


class Service(Enum):
    iam = "iam"
    registry = "registry"
    core = "core"


class ResourceType(Enum):
    # Gateway
    data_system = "data_system"
    data_source = "data_source"
    data_product = "data_product"
    data_unit = "data_unit"
    output = "output"
    journal_note = "journal_note"
    notification = "notification"
    secret = "secret"  # noqa: S105
    spark = "spark"
    tag = "tag"

    # IAM
    account = "account"
    group = "group"
    policy = "policy"
    user = "user"
    token = "token"  # noqa: S105
    signature = "signature"
    statement = "statement"
    resource = "resource"
    principal = "principal"

    # Registry
    core = "core"
    product = "product"
    mesh = "mesh"

    # Storage
    postgres = "postgres"
    minio = "minio"

    # Superset
    dataset = "dataset"


class Action(Enum):
    """Action class base.

    When implementing IAM actions in a service, extend this class.
    """

    star = "*"

    # Gateway
    data_product_create = "data_product:create"
    data_product_browse = "data_product:browse"
    data_product_read = "data_product:read"
    data_product_manage = "data_product:manage"

    data_source_create = "data_source:create"
    data_source_browse = "data_source:browse"
    data_source_read = "data_source:read"
    data_source_manage = "data_source:manage"

    data_system_create = "data_system:create"
    data_system_browse = "data_system:browse"
    data_system_read = "data_system:read"
    data_system_manage = "data_system:manage"

    data_unit_create = "data_unit:create"
    data_unit_browse = "data_unit:browse"
    data_unit_read = "data_unit:read"
    data_unit_manage = "data_unit:manage"

    notification_read = "notification:read"
    notification_manage = "notification:manage"

    output_create = "output:create"
    output_browse = "output:browse"
    output_read = "output:read"
    output_manage = "output:manage"

    secret_create = "secret:create"  # noqa: S105
    secret_browse = "secret:browse"  # noqa: S105
    secret_read = "secret:read"  # noqa: S105
    secret_manage = "secret:manage"  # noqa: S105

    tag_create = "tag:create"
    tag_browse = "tag:browse"
    tag_manage = "tag:manage"

    # Special action for creating manual spark jobs for development tests.
    spark_develop = "spark:develop"

    journal_note_manage = "journal_note:manage"

    # IAM
    self = "self"  # Allow a user to perform an action on themselves

    account_create = "account:create"
    account_delete = "account:delete"
    account_manage = "account:manage"
    account_member = "account:member"

    user_create = "user:create"
    user_delete = "user:delete"
    user_manage = "user:manage"
    user_browse = "user:browse"

    group_create = "group:create"
    group_browse = "group:browse"
    group_read = "group:read"
    group_manage = "group:manage"

    policy_create = "policy:create"
    policy_browse = "policy:browse"
    policy_read = "policy:read"
    policy_manage = "policy:manage"

    key_manage = "key:manage"

    validate = "auth:validate"

    statement_browse = "statement:browse"
    resource_browse = "resource:browse"

    principal_browse = "principal:browse"

    # Registry
    core_announce = "core:announce"
    core_register = "core:register"
    core_migrate = "core:migrate"  # REMOVE
    core_manage = "core:manage"
    core_access = "core:access"
    _core_browse = "core:browse"  # DEPRECATED

    product_register = "product:register"
    product_update = "product:update"
    product_remove = "product:remove"

    product_consume = "product:consume"
    product_subscribe = "product:subscribe"

    # Storage
    postgres_read = "postgres:read"
    postgres_write = "postgres:write"
    postgres_manage = "postgres:manage"
    minio_read = "minio:read"
    minio_write = "minio:write"
    minio_manage = "minio:manage"

    # Superset
    dataset_read = "dataset:read"

    # Service
    service_core = "service:core"


class EffectEnum(Enum):
    """Default effect enum for use with IAM actions."""

    allow = "allow"
    deny = "deny"


ServiceResource = [
    # Gateway
    (Service.core, ResourceType.data_system),
    (Service.core, ResourceType.data_source),
    (Service.core, ResourceType.data_product),
    (Service.core, ResourceType.data_unit),
    (Service.core, ResourceType.output),
    (Service.core, ResourceType.journal_note),
    (Service.core, ResourceType.notification),
    (Service.core, ResourceType.secret),
    (Service.core, ResourceType.spark),
    (Service.core, ResourceType.tag),
    # IAM
    (Service.iam, ResourceType.account),
    (Service.iam, ResourceType.group),
    (Service.iam, ResourceType.policy),
    (Service.iam, ResourceType.principal),
    (Service.iam, ResourceType.resource),
    (Service.iam, ResourceType.signature),
    (Service.iam, ResourceType.statement),
    (Service.iam, ResourceType.token),
    (Service.iam, ResourceType.user),
    # Registry
    (Service.registry, ResourceType.core),
    (Service.registry, ResourceType.product),
    (Service.registry, ResourceType.mesh),
    # Storage
    (Service.core, ResourceType.postgres),
    (Service.core, ResourceType.minio),
    # Superset
    (Service.core, ResourceType.dataset),
]


RL = t.TypeVar("RL", bound="ResourceLike")


@dataclass
class ResourceLike:
    all_: str | None = None

    xrn: str = "urn"
    partition: str = ""
    service: str = ""
    identifier: str = ""
    account_id: str = ""
    resource_type: str = ""
    sub_type: str = ""
    resource_id: str | None = None

    PATTERN_RULE = "[a-zA-Z0-9_\\-]{1,50}"
    OPTIONAL_PATTERN_RULE = "[a-zA-Z0-9_\\-]{0,50}"
    SUB_RESOURCE_TYPE_RULE = "[a-z-]{1,50}"
    ADDITIONAL_PATTERN_RULE = r"(\*|[a-zA-Z0-9_\-]{1,50})"
    # Valid are:
    # - *
    # - urn or nrn
    # - urn:partition:service:identifier:account_id:resource_type
    # - urn:partition:service:identifier:account_id:resource_type:*
    # - urn:partition:service:identifier:account_id:resource_type:resource_id
    # - urn:partition:service:identifier:account_id:resource_type:sub-type:resource_id
    #
    # Invalid are:
    # Pathlike
    # - urn:partition:service:identifier:account_id:resource_type/resource_id
    # - urn:partition:service:identifier:account_id:resource_type:resource_id/sub/path
    # - urn:partition:service:identifier:account_id:resource_type:resource_id:sub:path
    rule = PATTERN_RULE
    optional_rule = OPTIONAL_PATTERN_RULE
    sub_rule = SUB_RESOURCE_TYPE_RULE
    additional_rule = ADDITIONAL_PATTERN_RULE
    RESOURCE_PATTERN = rf"^((?P<all_>\*)|(?P<xrn>[un]rn):(?P<partition>{rule}):(?P<service>{rule}):(?P<identifier>{optional_rule})?:(?P<account_id>{rule})?:(?P<resource_type>{rule})(([:](?P<sub_type>{sub_rule}))?[:](?P<resource_id>{additional_rule}))?)$"  # noqa: E501

    @classmethod
    def parse(cls: type[RL], s: str) -> RL:
        m = re.match(cls.RESOURCE_PATTERN, s)
        if not m:
            msg = f"Could not parse the resource {s}"
            raise error.InvalidResourceFormatError(msg)
        m_dict = m.groupdict()
        return cls(
            all_=m_dict["all_"],
            xrn=m_dict["xrn"] or "",
            partition=m_dict["partition"] or "",
            service=m_dict["service"] or "",
            identifier=m_dict["identifier"] or "",
            account_id=m_dict["account_id"] or "",
            resource_type=m_dict["resource_type"] or "",
            sub_type=m_dict["sub_type"] or "",
            resource_id=m_dict["resource_id"],
        )

    @property
    def urn(self) -> str:
        if self.all_:
            return self.all_

        full_xrn = (
            f"{self.xrn}:{self.partition}:{self.service}:{self.identifier}:{self.account_id}:{self.resource_type}"
        )
        if self.resource_id:
            resource_id = self.resource_id if not self.sub_type else f"{self.sub_type}:{self.resource_id}"
            full_xrn = f"{full_xrn}:{resource_id}"
        return full_xrn


RR = t.TypeVar("RR", bound="ResourceReader")


@dataclass
class ResourceReader(ResourceLike):
    STAR = "*"

    @classmethod
    def to_root(cls: type[RR], other: RR) -> RR:
        return cls(
            all_=other.all_,
            xrn=other.xrn,
            partition=other.partition,
            service=other.service,
            identifier=other.identifier,
            account_id="root",
            resource_type=other.resource_type,
            sub_type=other.sub_type,
            resource_id=other.resource_id,
        )

    def to_list(self) -> list[str]:
        if self.all_ is not None and self.is_any():
            return [self.all_]

        ret = [
            el or ""
            for el in [
                self.xrn,
                self.partition,
                self.service,
                self.identifier,
                self.account_id,
                self.resource_type,
                self.sub_type,
                self.resource_id,
            ]
        ]
        if not self.sub_type:
            ret.pop(6)

        return ret

    def is_any(self) -> bool:
        return self.all_ is not None and self.all_ == self.STAR

    def is_any_resource_id(self) -> bool:
        if self.resource_id == self.STAR:
            return True
        return False

    def __getitem__(self, item: slice | int) -> str:
        """Get a specific item or subset of items from a resource.

        Examples:
        -------
        Resource("a:b:c:d")[0] -> "a"
        Resource("a:b:c:d")[1:3] -> "b:c"
        """
        if isinstance(item, slice):
            if item.stop == -1 and self.sub_type:
                item = slice(item.start, -2)
            return ":".join(self.to_list()[item]).rstrip(":")
        return self.to_list()[item]

    def __len__(self) -> int:
        """Return length of resource elements."""
        return len(self.to_list())


RB = t.TypeVar("RB", bound="ResourceBase")


@dataclass
class ResourceBase(abc.ABC, ResourceLike):
    """Resource class contains information about resource.

    Args:
    ----
        partition (str): geographic location of the system
        service (str): name of the service, for the core-gateway it is fixed to the core
        identifier (t.Union[str, None]): identifier of the core
        account_id (t.Union[str, None]): owner id of the core
        resource_type (str): name of the resource type
        resource_id (t.Union[str, None]): identificator of the resource, can have a nested structure
    """

    @classmethod
    def generate(
        cls: type[RB],
        xrn: str = "urn",
        partition: str = "partition",
        service: str = "service",
        identifier: str = "identifier",
        account_id: str = "owner",
        resource_type: str = "resource_type",
        sub_type: str = "",
        resource_id: str | tuple | None = None,
        all_: str | None = None,
    ) -> RB:
        """Generate Resource."""
        init_kwargs = {
            "xrn": xrn,
            "partition": partition,
            "service": service,
            "identifier": identifier or "",
            "account_id": account_id or "",
            "resource_type": resource_type,
            "sub_type": sub_type or "",
            "all_": all_,
        }

        if isinstance(resource_id, str):
            init_kwargs["resource_id"] = resource_id
        elif isinstance(resource_id, tuple):
            init_kwargs["resource_id"] = cls.format_resource_id(*resource_id)

        return cls(**init_kwargs)

    @property
    def urn_template(self) -> str:
        if self.all_:
            return self.all_

        full_xrn = (
            f"{self.xrn}:{self.partition}:{self.service}:{self.identifier}:{self.account_id}:{self.resource_type}"
        )
        if self.resource_id:
            resource_id = "{resource_id}" if not self.sub_type else f"{self.sub_type}:{{resource_id}}"
            full_xrn = f"{full_xrn}:{resource_id}"
        return full_xrn

    @classmethod
    @abc.abstractmethod
    def get_resource_id_template(cls) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def format_resource_id(cls, *args) -> str:
        pass



import json
import typing

import keycloak


class KeycloakClient:
    """Keycloak client that provides routes to the Keycloak."""

    def __init__(self, host: str, realm: str, client_id: str, client_secret: str) -> None:
        self._host = host
        self._realm = realm
        self._client_id = client_id
        self._client_secret = client_secret

    @property
    def openid(self) -> keycloak.KeycloakOpenID:
        """Returns the openid client for the Keycloak."""
        return keycloak.KeycloakOpenID(
            server_url=f"{self._host}/auth",
            client_id=self._client_id,
            realm_name=self._realm,
            client_secret_key=self._client_secret,
        )

    def introspect_token(self, token: str) -> dict:
        """Retrieve introspect token from keycloak.

        Useful to validat that a user token is still active.
        """
        return self.openid.introspect(token)

    def get_token_by_password(self, user: str, password: str) -> dict:
        """Retrieve access_token with user credentials."""
        return self.openid.token(user, password)

    def refresh_token(self, refresh_token: str) -> dict:
        """Obtain a new access_token using a refresh_token."""
        return self.openid.refresh_token(refresh_token)

    def logout_user(self, refresh_token: str) -> dict:
        """Logout of keycloak session."""
        return self.openid.logout(refresh_token)

    def userinfo(self, token: str) -> dict:
        """Get user info for supplied access_token."""
        return self.openid.userinfo(token)

    def decode_token(self, any_token: str, key: str) -> dict:
        return self.openid.decode_token(any_token, key)

    @staticmethod
    def parse_error(keycloak_error: keycloak.KeycloakError) -> str | dict[str, typing.Any]:
        try:
            return json.loads(keycloak_error.error_message)
        except json.decoder.JSONDecodeError:
            return str(keycloak_error.error_message)
